//
//  DBAPI2BaseTask.m
//
//

#import "DBAPI2BaseTask.h"

#import "KPSharedNetworkActivityIndicator.h"

#import "NSError+LoadingErrorClassification.h"

#import "KPLogger.h"

#if !__has_feature(objc_arc)
#error This class requires automatic reference counting
#endif

@interface DBAPI2BaseTask ()

@property (nullable, nonatomic, strong) DBRequestError *requestError;

@property (nonatomic, strong) dispatch_semaphore_t internalTaskProcessingSemaphore;
@property (nonatomic, strong) NSMutableArray<DBTask *> *internalDbTasks;

@property (nonatomic, assign) BOOL isCancelled;

@end

@implementation DBAPI2BaseTask

- (nullable DBFILESUserAuthRoutes *)fileAPI {
    
    return [DBClientsManager authorizedClient].filesRoutes;
}

- (void)registerInternalDbTask:(nonnull DBTask *)internalDbTask {
    
    @synchronized (self) {
    
        if (!internalDbTask) {
            return;
        }
        
        if (!_internalDbTasks) {
            _internalDbTasks = [NSMutableArray new];
        }
        
        [_internalDbTasks addObject:internalDbTask];
    }
}

- (void)cancel {
 
    KPLog(@"");
    
    @synchronized (self) {
    
        [self cancelInternalTasks];
        
        _isCancelled = YES;
    }
}

- (void)cancelInternalTasks {
    
    for (DBTask *task in _internalDbTasks) {
        
        KPLog(@"[dbTask: %@]", task);
        
        [task cancel];
    }
    
    [_internalDbTasks removeAllObjects];
}

- (BOOL)doSynchronously {
    
    KPLog(@"");
    
    if (![self fileAPI]) {
        
        KPLogWarn(@"File API is not available...skipping");
        
        return NO;
    }
    
    BOOL shouldShowActivityIndicator = [self shouldShowActivityIndicator];
    
    if (shouldShowActivityIndicator) {
        [[KPSharedNetworkActivityIndicator sharedInstance] incrementActivityCount];
    }
    
    BOOL done = [self doInternalWorkSynchronously];
    
    if (shouldShowActivityIndicator) {
        [[KPSharedNetworkActivityIndicator sharedInstance] decrementActivityCount];
    }
    
    return done;
}

- (BOOL)doInternalWorkSynchronously {
    
    BOOL done = NO;
    
    do
    {
        done = [self doInternalTaskSynchronously];
        
        KPLog(@"[requestError: %@]", _requestError);
        
        if ([_requestError isRateLimitError]) {
            
            [self waitBeforeMakingAnyAdditionalRequests];
        }
        else if ([_requestError isAuthError]) {
            
            break;
        }
        else {
            
            // Skip processing of other types the request error.
            break;
        }
        
    } while (!_isCancelled);
    
    return done;
}

- (BOOL)doInternalTaskSynchronously {
    
    _internalTaskProcessingSemaphore = dispatch_semaphore_create(0);
    
    [self cleanupOutput];
    
    DBTask *internalDbTask = [self doInternalTaskAsynchronously];
    
    if (internalDbTask) {
    
        [self registerInternalDbTask:internalDbTask];
        
        dispatch_semaphore_wait(_internalTaskProcessingSemaphore, DISPATCH_TIME_FOREVER);
    }
    else {
   
        KPLog(@"*** internalDbTask was not created ***");
        
        [self stopWaiting];
    }
    
    return [self returnResult];
}

- (BOOL)shouldShowActivityIndicator {
    
    // YES - BY DEFAULT
    
    return YES;
}

- (void)cleanupOutput {
    
    // MUST BE OVERRIDDEN
    
    _requestError = nil;
}

- (BOOL)returnResult {
    
    // MUST BE OVERRIDDEN
    return NO;
}

- (nullable DBTask *)doInternalTaskAsynchronously {
    
    // MUST BE OVERRIDDEN
    return nil;
}

- (void)logWorkResults {
    
    // MUST BE OVERRIDDEN
}

- (void)updateRequestError:(nullable DBRequestError *)error {
    
    _requestError = error;
}

- (void)stopWaiting {
    
    [self logWorkResults];
    
    dispatch_semaphore_signal(_internalTaskProcessingSemaphore);
}

#pragma mark -
#pragma mark Processing common network errors
#pragma mark -

- (void)waitBeforeMakingAnyAdditionalRequests {
    
    // Seconds to wait before making any additional requests in the event of a rate-limit error.
    NSInteger backOff = [_requestError.backoff intValue];
    
    [NSThread sleepForTimeInterval:backOff];
}

- (void)refreshAccessToken {
    
    dispatch_semaphore_t refreshTokenSemaphore = dispatch_semaphore_create(0);
    
    DBOAuthManager *dbOAuthManager = [DBOAuthManager sharedOAuthManager];
    
    DBAccessToken *accessToken = [dbOAuthManager retrieveFirstAccessToken];
    
    [dbOAuthManager refreshAccessToken:accessToken
                                scopes:@[]
                                queue:nil
                            completion:^(DBOAuthResult * __nullable authResult) {
        
        dispatch_semaphore_signal(refreshTokenSemaphore);
    
        KPLog(@"[authResult: %@]", authResult);
    }];
    
    dispatch_semaphore_wait(refreshTokenSemaphore, DISPATCH_TIME_FOREVER);
}

@end
