//
//  KPDBSyncContentChangesObserver.m
//
//

#ifdef DROPBOX_FEATURE

#import "KPDBSyncContentChangesObserver.h"

#import "KPDBLocalFileSystemObserver.h"

#import "DBSyncFolder+Creating.h"
#import "NSOperationQueue+Utils.h"
#import "NSError+LoadingErrorClassification.h"

#import "KPLogger.h"

#import "DBAPI2WaitForChangesInDBFolderTask.h"

#if !__has_feature(objc_arc)
#error This class requires automatic reference counting
#endif

// In seconds
static NSTimeInterval const WAIT_TIMEOUT = (1 * 60.0);

@interface KPDBSyncContentChangesObserver () <KPDBLocalFileSystemObserverDelegate>

@property (nonatomic, weak) id<IKPDBSyncContentChangesObserverDelegate> delegate;

@property (nonatomic, strong) KPDBSyncContentSnapshot *snapshot;
@property (nonatomic, assign) BOOL observeDBChanges;
@property (nonatomic, assign) BOOL observeLocalChanges;

@property (nonatomic, strong) NSOperationQueue *notificationQueue;
@property (nonatomic, strong) NSOperationQueue *operationQueue;
@property (nonatomic, assign) BOOL cancelled;

@property (nonatomic, strong) KPDBLocalFileSystemObserver *localFileSystemObserver;

@property (nonatomic, strong) DBAPI2WaitForChangesInDBFolderTask *waitForSharedDbChangesTask;
@property (nonatomic, strong) DBAPI2WaitForChangesInDBFolderTask *waitForUniqueDbChangesTask;

@end

@implementation KPDBSyncContentChangesObserver

#pragma mark -
#pragma mark Life cycle
#pragma mark -

+ (nonnull instancetype)observeSyncContentWithSnapshot:(KPDBSyncContentSnapshot *)snapshot
                                      observeDBChanges:(BOOL)observeDBChanges
                                   observeLocalChanges:(BOOL)observeLocalChanges
                        notificationQueue:(nullable NSOperationQueue *)notificationQueue
                            delegate:( id<IKPDBSyncContentChangesObserverDelegate>)delegate {
    
    KPDBSyncContentChangesObserver *observer =
        [[KPDBSyncContentChangesObserver alloc] initWithSnapshot:snapshot
                                                observeDBChanges:observeDBChanges
                                             observeLocalChanges:observeLocalChanges
                                               notificationQueue:notificationQueue
                                                        delegate:delegate];
    
    [observer start];
    
    return observer;
}

- (nonnull instancetype)initWithSnapshot:(nonnull KPDBSyncContentSnapshot *)snapshot
                        observeDBChanges:(BOOL)observeDBChanges
                     observeLocalChanges:(BOOL)observeLocalChanges
                       notificationQueue:(nullable NSOperationQueue *)notificationQueue
                                delegate:(nonnull id<IKPDBSyncContentChangesObserverDelegate>)delegate {
    
    if (self = [super init]) {
        
        _delegate = delegate;
        _snapshot = snapshot;
        _observeDBChanges = observeDBChanges;
        _observeLocalChanges = observeLocalChanges;
        
        if (!notificationQueue) {
            _notificationQueue = [NSOperationQueue mainQueue];
        } else {
            _notificationQueue = notificationQueue;
        }
    }
    
    return self;
}

#pragma mark -
#pragma mark Public operations
#pragma mark -

- (void)cancel {

    KPLog(@"");
    
    _delegate = nil;
    _cancelled = YES;
    
    [_localFileSystemObserver stop];
    [_waitForSharedDbChangesTask cancel];
    [_waitForUniqueDbChangesTask cancel];
}

#pragma mark -
#pragma mark Private
#pragma mark -

- (void)start {
    
    KPLog(@"");
    
    if (!_operationQueue) {
        _operationQueue = [NSOperationQueue concurrentServiceUtilityQueueWithOwner:self];
    }
    
    weakify(self);
    [_operationQueue addOperationWithBlock:^{
        
        strongify(self);
        
        if (self.delegate && !self.cancelled) {

            if (self.observeLocalChanges) {
                
                [self startWaitingForLocalChanges];
            }
            
            if (self.observeDBChanges) {
                
                self.waitForSharedDbChangesTask =
                    [self startWaitingForDBChangesWithCursor:
                        self.snapshot.dbSharedContentCursor];
                
                self.waitForUniqueDbChangesTask =
                    [self startWaitingForDBChangesWithCursor:
                        self.snapshot.dbUniqueContentCursor];
            }
        }
    }];
}

- (NSString *)uniqueFolderPath {
    
    return [[DBSyncFolder folderFromSettings] fullPath2LocalUniqueFolder];
}

- (void)localFileSystemDidChange {
    
    KPLog(@"");
    
    [_delegate contentObserverDidDetectLocalChanges:self];
}

- (void)startWaitingForLocalChanges {
    
    KPLog(@"");
    
    if (!_localFileSystemObserver) {
        
        _localFileSystemObserver =
            [[KPDBLocalFileSystemObserver alloc] initWithPath4Observing:[self uniqueFolderPath]
                                                               delegate:self];
    }
    
    [_localFileSystemObserver start];
}

- (nullable DBAPI2WaitForChangesInDBFolderTask *)
    startWaitingForDBChangesWithCursor:(NSString *)contentCursor {
    
    KPLog(@"[cursor: %@]", contentCursor);
    
    if (!contentCursor) {
        return nil;
    }
    
    DBAPI2WaitForChangesInDBFolderTask *task =
        [[DBAPI2WaitForChangesInDBFolderTask alloc] initWithCursor:contentCursor
                                                           timeout:WAIT_TIMEOUT];
    
    weakify(self);
    [_operationQueue addOperationWithBlock:^{
        
        strongify(self);
        
        [self waitForDBCursor:contentCursor task:task];
    }];
    
    return task;
}

- (void)waitForDBCursor:(nonnull NSString *)contentCursor
                   task:(DBAPI2WaitForChangesInDBFolderTask *)task {
    
    do
    {
        @autoreleasepool {
            
            [task doSynchronously];
            
            if (!self.delegate || self.cancelled) {
            
                KPLogWarn(@"[delegate: %@][cancelled: %d]", self.delegate, self.cancelled);
                
                break;
            }
            
            if (task.changes) {
                
                KPLogWarn(@"[task.changes: %d]", task.changes);
                
                [[self notificationQueue] addOperationWithBlock:^{
                    [self.delegate contentObserverDidDetectDBChanges:self];
                }];
                
                break;
            }

            KPLog(@"[task.routeError: %@][task.requestError: %@]", task.routeError, task.requestError);
            
            KPLog(@"Continue waiting for: [cursor: %@]", contentCursor);

            [task cancel];
        }
        
    } while (YES);
    
    [task cancel];
    
    KPLog(@"Waiting finished:[cursor: %@]", contentCursor);
}

@end

#endif
